/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  control.c -- forward and parse the control stream.
  
  ***************************************/


#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include "control.h"
#include "common.h"

#ifdef USE_CACHE
#include "cache.h"
#endif

extern session_info *info;

int extract_clientcmd(unsigned char * buf, unsigned char * cmd, unsigned char * arg);
int extract_servercmd(unsigned char * buf, int * code, unsigned char * msg);

void parse_client_cmd(unsigned char *cmd, unsigned char * arg);
void parse_server_cmd(int code, unsigned char * msg);

void pasv_parse(const unsigned char *cmd, const unsigned char *arg);
void port_parse(const unsigned char *cmd, const unsigned char * arg);
void pasv_reply(const unsigned char *msg);

void write_all(int fd, unsigned char *buf);

/* ------------------------------------------------------------- **
** Forward control stream data from client -> server. Non blocking 
** ------------------------------------------------------------- */
void client_control_forward()
{
	int i, j;
	unsigned char cmdbuf[5], argbuf[MAX_LINE_LEN];

	do{
		ioctl(info->client_control.fd, FIONREAD, &i);
	}while(i<0 && errno==EINTR);

	if(i<1) quit(CLIENT_CLOSE);

	j=strlen(info->client_control.buf);
	i = iread(info->client_control.fd, info->client_control.buf+j,
		 (i > (BUF_LEN-1) - j ? (BUF_LEN-1) - j : i) );
	info->client_control.buf[i+j]=0;

	/*buf is NULL terminated, but could contain anything*/
					 
	while( (i=extract_clientcmd(info->client_control.buf, cmdbuf, argbuf)) == 0 )
		parse_client_cmd(cmdbuf, argbuf);

	if(i<0) {
		send_message(500,"You're sending me rubbish. Goodbye!");
		quit(CLIENT_RUBBISH);
	}
}

/* ------------------------------------------------------------- **
** Forward control stream data from server -> client. Non blocking 
** ------------------------------------------------------------- */
void server_control_forward()
{
	int i, j, code;
	unsigned char argbuf[MAX_LINE_LEN];
	
	do{
		ioctl(info->server_control.fd, FIONREAD, &i);
	}while(i<0 && errno==EINTR);

	if(i<1) quit(SERVER_CLOSE);

	j=strlen(info->server_control.buf);

	i = iread(info->server_control.fd, info->server_control.buf+j, 
		 (i > (BUF_LEN-1)-j ? (BUF_LEN-1)-j : i) );
	info->server_control.buf[i+j]=0;

	/*buf is NULL terminated, but could contain anything*/

        while( (i=extract_servercmd(info->server_control.buf, &code, argbuf)) == 0 )
		parse_server_cmd(code,argbuf);

        if(i<0) {
		send_message(500,"FTP server is sending you rubbish! Closing connection.");
		quit(SERVER_RUBBISH);
	}
}

/***************************************************************************
 *
 * Functions which read the raw control stream --- be careful
 *
 **************************************************************************/

/* ------------------------------------------------------------- **
** Removes one line of control stream from buf (up to '\r\n'). <= 5
** chars are returned in cmd (up to 4 + \0), and <= MAX_LINE_LEN in
** arg. Must accept any NULL terminated buf and give sane output.
**  
** returns 0 on success, 1 if there is not a complete line in
** buf, and -X on non-sane buf. contents of buf unchanged on
** return(1), undefined on return(-X)
** ------------------------------------------------------------- */
int extract_clientcmd(unsigned char *buf, unsigned char *cmd, unsigned char *arg)
{
	unsigned char *p,*cmdend,*argstart,*end;

	cmdend=argstart=end=p=strpbrk(buf,"\r\n \t");
	if(p==NULL) return(1);

	if(*p==' ' || *p=='\t'){
		for(argstart=p;*argstart==' ' || *argstart=='\t';argstart++);
		end=p=strpbrk(p, "\r\n");
	}
	if(p==NULL) return(1);
	if(*p=='\r')
		if( (end=strpbrk(p,"\n"))==NULL ) return(1);
	
	/* We have a complete command. Start copying it over.
	   eg. "USER   ftp\r\r\r\n"  (Microsoft servers don't do \r\n !)
	   buf__|   |  |  |     |__end
        cmdend______|  |  |______p
	argstart_______|               */

	*cmdend=0;
	*p=0;

	if(buf[0]==IAC) do_linecontrol(buf);  /* Telnet cmd */

	if(strlen(buf)>4 || strlen(argstart) > MAX_LINE_LEN-1 ) {
		debug2("Client sent us \"%s\", ",buf);
		debug2("\"%s\"\n",argstart);
		return(-1);
	}
	strcpy(cmd, buf);
	strcpy(arg,argstart);
	
	memmove(buf, end+1, strlen(end+1)+1);
	return(0);
}

/* ------------------------------------------------------------- **
** As extract_clientcmd, but returns the code as an int.
** If code==0 then we are in a multiline. code<0 means we are
** starting a multiline. 
** ------------------------------------------------------------- */
int extract_servercmd(unsigned char *buf, int * code, unsigned char *msg)
{
	static int multiline=0;
	int i;
	unsigned char *p,*end;

	end=p=strpbrk(buf,"\r\n");

	if(p==NULL) return(1);
	if(*p=='\r')
		if( (end=strpbrk(p,"\n"))==NULL ) return(1);
	
	/* We have a complete command. Start copying it over.
	   eg. "501  go away now\r\r\n"  (Microsoft servers don't do \r\n !)
	   buf__|               |   |__end
	                        |______p       */

	*p=0;

	if(multiline == 0){
		for(i=0;i<3;i++) if( buf[i] < '0' || buf[i] > '9') return(-1);
		if(buf[3]!=' ' && buf[3]!='-') return(-1);
		*code=atoi(buf);
		if(buf[3] == '-') {  /* Start of a multiline */

			multiline = *code;
			*code=-*code;
		}
		if(strlen(buf+4) > MAX_LINE_LEN-1 ) return(-1);
		strcpy(msg,buf+4);
	} else {
		if(isdigit(buf[0]) && isdigit(buf[1]) && isdigit(buf[2])
		   && buf[3]==' ' && (*code=atoi(buf)) == multiline ) { /* End multiline */
			multiline=0;
			if(strlen(buf+4) > MAX_LINE_LEN-1) return(-1);
			strcpy(msg,buf+4);
		} else {             /* Continue Multiline */
			*code=0;
			if(strlen(buf) > MAX_LINE_LEN-1) return(-1);
			strcpy(msg,buf);
		}
	}

	memmove(buf, end+1, strlen(end+1)+1);
	return(0);
}

/* ------------------------------------------------------------- **
** Functions to parse commands. 
** ------------------------------------------------------------- */
void parse_client_cmd(unsigned char *cmd,unsigned char *arg)
{

#if (CHECK_LEVEL == 1) /* Check for dodgy characters */
        int i;

	for(i=0;i<strlen(cmd);i++)
		if(cmd[i]>126 || cmd[i]<32) cmd[i]='?';
	for(i=0;i<strlen(arg);i++)
		if(arg[i]>126 || arg[i]<32) arg[i]='?';

#endif

#ifdef USE_CACHE
	if(config_usecache()){
		if(cache_parse_cmd(cmd, arg)) {
			debug2("cache.c dealt with %s\n",cmd);
			return;
		}
	}
#endif

	if(!strncasecmp(cmd,"PASV",4)) pasv_parse(cmd,arg);
	else if(!strncasecmp(cmd,"PORT",4)) port_parse(cmd, arg);
	else if(!strncasecmp(cmd,"ABOR",4)) {
		unsigned char tmp[3];
		tmp[0]=IAC;
		tmp[1]=IP;
		tmp[2]=IAC;
		send(info->server_control.fd,tmp,3,MSG_OOB);
		tmp[0]=DM;
		send(info->server_control.fd,tmp,1,0);
		send_command(cmd, arg);
	}
	else send_command(cmd, arg);
}

void parse_server_cmd(int code, unsigned char * msg)
{
	int i;

	for(i=0;i<strlen(msg);i++)
		if(msg[i]>126 || msg[i]<32) msg[i]='?';

	/*FIXME What was this supposed to do? */
	if(code==0) {  /* Part of multiline */
		send_message(code, msg);
		return;
	}

	switch(code)
	{
	case 227:
		pasv_reply(msg);
		break;
#ifdef USE_CACHE
	case 257:
		cache_pwd_reply(msg);
		break;
#endif
	default:
		send_message(code, msg);
	}
}

/* ------------------------------------------------------------- **
** Parse the 227 reply message. Store the port it gives, open
** one of our own, and rewrite the 227 command with that.  
** ------------------------------------------------------------- */
void pasv_reply(const unsigned char * msg)
{
	unsigned char newbuf[60];
	int a1,a2,a3,a4,p1,p2;
	address_info tmp;

	info->server_data.address=extract_address(msg);
	if(info->server_data.address.addr==0) {
		send_message(500, "Bad passive command from server");
		return;
	}

	if(config_usesameaddress()){
		if(info->server_data.address.addr != info->server_control.address.addr){
			send_message(500,"Bad PORT command");
			return;
		}
	}


        /* Get the address of the local interface. We can't look at
        ** client_control.fd for this as we are doing transparent
        ** proxying. This piece of code ought to be taken out and shot 
	**  (yes - it opens a UDP port and then closes it again) */

	tmp.addr=config_getaddress();
	if(tmp.addr==0) {
		struct sockaddr_in stuff;
		socklen_t namelen;
		int sockfd;
		
		if( (sockfd = socket(AF_INET, SOCK_DGRAM, 0))==-1) {
			perror("socket");
			exit(1);
		}
		stuff.sin_family      = AF_INET;
		stuff.sin_addr.s_addr = info->client_control.address.addr;
		stuff.sin_port        = htons(12345);
		namelen = sizeof(stuff);
			
		if(connect(sockfd, (struct sockaddr *)&stuff, namelen)==-1){
			write_log(ERROR, "Unable to determine local interface address");
			exit(1);
		}
		
		getsockname(sockfd, &stuff, &namelen);
	}
	tmp.port=0;	
	
	info->listen=listen_on_socket(&tmp);

	if(info->listen == -1) {
                send_message(451, "Screwed up pasv command.");
                return;
        } else info->state |= LISTEN;		

        n2com(tmp,&a1,&a2,&a3,&a4,&p1,&p2);

        snprintf(newbuf,60,"Entering Passive Mode (%d,%d,%d,%d,%d,%d)",
		 a1,a2,a3,a4,p1,p2);

        debug2("  Rewritten 227 command to:\n     \033[34m%s\033[37m\n",newbuf);

        send_message(227, newbuf);
	info->state &= ~ACTIVE; /* FIXME: Do we want this here? */
} 

/* ------------------------------------------------------------- **
** Parse the PORT command in arg Store the client's data listening
** port, open one of our own, and send this to the server in a
** rewritten PORT command.
** ------------------------------------------------------------- */
void port_parse(const unsigned char * cmd, const unsigned char *arg)
{
	unsigned char newbuf[40];
	int a1,a2,a3,a4,p1,p2;
	address_info tmp;
        struct sockaddr_in listenaddr;
        socklen_t len;

	info->client_data.address=extract_address(arg);
	if(info->client_data.address.addr==0){
		send_message(500,"Bad PORT command");
		return;
	}

	if(config_usesameaddress()){
		if(info->client_data.address.addr != info->client_control.address.addr){
			send_message(500,"Bad PORT command");
			return;
		}
	}
		
	if(config_bouncedefend()){
		if(ntohs(info->client_data.address.port)<1024){
			send_message(500,"Bad PORT command");
			return;
		}
	}

	info->state |= ACTIVE;
	close(info->listen);
	info->state &= ~LISTEN;

        len=sizeof(listenaddr);
        getsockname(info->server_control.fd, &listenaddr, &len);

	tmp.addr=listenaddr.sin_addr.s_addr;
	tmp.port=0;
	info->listen=listen_on_socket(&tmp);

	if(info->listen == -1){
		send_message(451, "Proxy unable to comply.");
		return;
	} else info->state |= LISTEN;

	n2com(tmp,&a1,&a2,&a3,&a4,&p1,&p2);

	snprintf(newbuf,40,"%d,%d,%d,%d,%d,%d",a1,a2,a3,a4,p1,p2);

	debug2("  Rewritten PORT command to:\n     \033[31m%s\033[37m\n",newbuf);

	send_command(cmd, newbuf);
}

/* ------------------------------------------------------------- **
** Intercepted a PASV command. 
** ------------------------------------------------------------- */
void pasv_parse(const unsigned char *cmd, const unsigned char *arg)
{
	debug("  Intercepted a PASV command\n");

	info->state &= ~ACTIVE;
	close(info->listen);
	info->state &= ~LISTEN;

	send_command(cmd, arg );
}

/* ------------------------------------------------------------- **
** Send command to server 
** ------------------------------------------------------------- */
void send_command(const unsigned char * cmd, const unsigned char *arg )
{
	unsigned char buf[BUF_LEN];

	if(strlen(arg)==0)
		snprintf(buf,BUF_LEN-1, "%s\r\n",cmd);
	else    snprintf(buf,BUF_LEN-1,"%s %s\r\n",cmd, arg);

	write_all(info->server_control.fd, buf);

	debug2("  C: \033[31m%s\033[37m",buf); /*Ooooh Colour*/
}
		
/* ------------------------------------------------------------- **
** Write message to client. Don't give unchecked data!
** ------------------------------------------------------------- */
void send_message(int code, const unsigned char *msg)
{
	unsigned char buf[BUF_LEN];

	if(code>0) snprintf(buf,BUF_LEN-1,"%d %s\r\n",code, msg);
	else if(code<0) snprintf(buf,BUF_LEN-1,"%d-%s,\r\n",-code, msg); 
	else snprintf(buf, BUF_LEN-1, "%s\r\n", msg);

	write_all(info->client_control.fd, buf);

	debug2("  S: \033[34m%s\033[37m",buf);
	
}

/* ------------------------------------------------------------- **
** Write buf->fd. If fd has O_NONBLOCK set we will keep trying
** until TIMEOUT.
** ------------------------------------------------------------- */
void write_all(int fd, unsigned char *buf)
{
	fd_set writes;
	struct timeval timeout;
	unsigned char *p=buf;
	int i=0, tmp;
	time_t end;

	end=time(NULL)+config_timeout().tv_sec; /*Should we use another value?*/
	
	do{
		FD_ZERO(&writes);
		FD_SET(fd, &writes);
		if( (tmp=end-time(NULL))<0) break;
		timeout.tv_sec=tmp;
		timeout.tv_usec=0;

		i=iselect(FD_SETSIZE, NULL, &writes, NULL, &timeout);
		if(i<1) break;
		i=iwrite(fd, p,strlen(p));
		if(i<1) break;
		p+=i;
	} while(i<strlen(p));

	if(i<1){
		perror("writeall");
		write_log(ERROR, "Error writing to socket. Exiting");
		closeandexit(-1);
	}

}

	
