/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
ftp-proxy.c -- central code for managing forwarding.
               Also the data path forwarding stuff.
  ***************************************/


#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>

#include "ftp-proxy.h"
#include "control.h"

void run_proxy(void);
void setup_fds(fd_set *reads,fd_set *writes,struct timeval *timeout);
void acton_event(fd_set *reads,fd_set *writes);

void client_data_connect(void);
void server_data_connect(void);

void writebuf2client(void);
void writebuf2server(void);
void forwarddata2client(void);
void forwarddata2server(void);

void closedata(int closeme);

/* ------------------------------------------------------------- **
**  Sets up the control connection to the server and initialises
**  session wide parts of the info structure.
** ------------------------------------------------------------- */
void init_session(int fd, struct sockaddr_in source)
{
	struct sockaddr_in realdest;
	socklen_t scklen=sizeof(realdest);
	unsigned char tmp[MAX_LINE_LEN];

	info=(session_info *)malloc(sizeof(session_info));
	if(info==NULL) quit(MALLOC);
	memset(info,0,sizeof(session_info));

	/* Get the true destination */
	getsockname(fd, (struct sockaddr *)&realdest, &scklen);
	info->server_control.address.addr = realdest.sin_addr.s_addr;
	info->server_control.address.port = realdest.sin_port;

	info->client_control.address.addr = source.sin_addr.s_addr;
	info->client_control.address.port = source.sin_port;
	info->client_control.fd = fd;

	strcpy(tmp, addr2name(info->client_control.address.addr));
	write_log(INFO, "Connect from %s to %s", tmp,
		  addr2name(realdest.sin_addr.s_addr));

	debug("Connecting to server...");

	if( (info->server_control.fd=connect_to_socket(info->server_control.address)) ==-1){
		iwrite(info->client_control.fd,
		      "501 Proxy unable to contact ftp server\n",39);
		write_log(ERROR, "Connection closed -- unable to contact server");
		close(info->client_control.fd);
		exit(0);
	}

	debug("OK\n");

/*        fcntl(info->client_control.fd, F_SETFL, O_NONBLOCK);
        fcntl(info->server_control.fd, F_SETFL, O_NONBLOCK); */

	info->listen=info->client_data.fd=info->server_data.fd=-1;
	*info->server_control.buf=*info->client_control.buf=0;
	info->state=ACTIVE;

	run_proxy();
}

/* ------------------------------------------------------------- **
**  The main loop for each session.
** ------------------------------------------------------------- */
void run_proxy()
{
	fd_set reads, writes;
	struct timeval timeout;

	do{
		setup_fds(&reads,&writes,&timeout);

		if(iselect(FD_SETSIZE, &reads, &writes, NULL, &timeout)==0)
			quit(TIMEOUT);

		acton_event(&reads,&writes);
		
	} while(TRUE);	
}

/* ------------------------------------------------------------- **
**  Work out what we need to select on
** ------------------------------------------------------------- */
void setup_fds(fd_set *reads,fd_set *writes,struct timeval *timeout)
{
	FD_ZERO(reads);
	FD_ZERO(writes);
	
	FD_SET(info->client_control.fd,reads);
	FD_SET(info->server_control.fd,reads);
	
	if(info->state & SERVERDATA){
		if(info->client_data.buflen==0) 
			FD_SET(info->server_data.fd,reads);
		
		if(info->server_data.buflen!=0)
			FD_SET(info->server_data.fd,writes);
	}
	if(info->state & CLIENTDATA){
		if(info->server_data.buflen==0) 
			FD_SET(info->client_data.fd,reads);
		if(info->client_data.buflen!=0)
			FD_SET(info->client_data.fd,writes);
	}
	if(info->state & LISTEN) {
		FD_SET(info->listen,reads);
	}
	
	*timeout=config_timeout();
}


/* ------------------------------------------------------------- **
**  Examine the fd_sets, and act on them.
** ------------------------------------------------------------- */
void acton_event(fd_set *reads,fd_set *writes)
{

	if(FD_ISSET(info->client_control.fd,reads))
		client_control_forward();
	
	if(FD_ISSET(info->server_control.fd,reads))
		server_control_forward();
	
	if( (info->state & LISTEN) && FD_ISSET(info->listen,reads)){
		if(info->state & ACTIVE) server_data_connect();
		else client_data_connect();
	}
	
	if(info->state & SERVERDATA){
		if(FD_ISSET(info->server_data.fd, reads)) forwarddata2client();
		if(FD_ISSET(info->server_data.fd, writes)) writebuf2server();
	}
	if(info->state & CLIENTDATA){
		if(FD_ISSET(info->client_data.fd, reads)) forwarddata2server();
		if(FD_ISSET(info->client_data.fd, writes)) writebuf2client();
	}
}

/* ------------------------------------------------------------- **
**  Client just connected to the data line
** ------------------------------------------------------------- */
void client_data_connect()
{
        struct sockaddr_in client;
        int len = sizeof(client);

        debug("  Client has connected to proxy data line\n");
        info->client_data.fd=accept(info->listen, (struct sockaddr *)&client, &len);

	if(config_usesameaddress){
		if(client.sin_addr.s_addr!=info->client_control.address.addr){
			write_log(ATTACK, "Blocked %s from connecting to data line",
			       addr2name(client.sin_addr.s_addr));
			close(info->client_data.fd);
			return;
		}
	}

	if( !(info->state & SERVERDATA) ){
		info->server_data.fd=connect_to_socket(info->server_data.address);
		if(info->server_data.fd == -1) quit(NO_SERVER_DATA);
		info->server_data.buflen=0;
	}
	info->client_data.buflen=0;
        info->state |= CONNECTED;

        fcntl(info->client_data.fd, F_SETFL, O_NONBLOCK);
        fcntl(info->server_data.fd, F_SETFL, O_NONBLOCK);
}

/* ------------------------------------------------------------- **
**  Server just connected to the data line
** ------------------------------------------------------------- */
void server_data_connect()
{
	struct sockaddr_in server;
	int len=sizeof(server);
	
	debug("  Server has connected to proxy data line\n");
	info->server_data.fd=accept(info->listen, (struct sockaddr *) &server, &len);
	if(config_usesameaddress()) {
		if(server.sin_addr.s_addr!=info->server_control.address.addr){
			write_log(ATTACK, "Blocked %s from connecting to data line",
			       addr2name(server.sin_addr.s_addr));
			close(info->server_data.fd);
			return;
		}
	}
	info->client_data.fd=connect_to_socket(info->client_data.address);
	if(info->client_data.fd == -1) quit(NO_CLIENT_DATA);
	info->server_data.buflen=info->client_data.buflen=0;
	info->state |= CONNECTED;

        fcntl(info->client_data.fd, F_SETFL, O_NONBLOCK);
        fcntl(info->server_data.fd, F_SETFL, O_NONBLOCK);
}

/* ------------------------------------------------------------- **
** Read up to BUF_LEN, write what we can to the server without  
** blocking, and store The rest in the server data buffer. 
** ------------------------------------------------------------- */
void forwarddata2server()
{
	int i,j;
		
	i=iread(info->client_data.fd,info->server_data.buf,BUF_LEN-1);
	if(i<1) {
		closedata(CLIENTDATA);
		return;
	}
	j=iwrite(info->server_data.fd, info->server_data.buf, i);
	if(j==-1) j=0;
	debug2("    Forwarded %d/",j);
	debug2("%d bytes client --> server. ",i);
	memmove(info->server_data.buf, info->server_data.buf+j, i-j);
	info->server_data.buflen=i-j;
	debug2(" Server buffer length = %d\n",info->server_data.buflen);
}

/* ------------------------------------------------------------- **
** Read up to BUF_LEN, write what we can to the client without  
** blocking, and store The rest in the client data buffer. 
** ------------------------------------------------------------- */
void forwarddata2client()
{
	int i,j;

	i=iread(info->server_data.fd,info->client_data.buf,BUF_LEN-1);
	if(i<1) {
		closedata(SERVERDATA);
		return;
	}
	j=iwrite(info->client_data.fd, info->client_data.buf, i);
	if(j==-1) j=0;
	debug2("    Forwarded %d/",j);
	debug2("%d bytes server --> client.", i);
	memmove(info->client_data.buf, info->client_data.buf+j, i-j);
	info->client_data.buflen=i-j;
	debug2(" Client buffer length = %d\n",info->client_data.buflen);
}

/* ------------------------------------------------------------- **
**  Flush as much buffer as we can without blocking  
** ------------------------------------------------------------- */
void writebuf2client()
{
	int i;

	debug("    Writing buffer --> client\n");
	i=iwrite(info->client_data.fd,info->client_data.buf,info->client_data.buflen);
	if(errno==EPIPE){
		closedata(CLIENTDATA);
		return;
	}
	memmove(info->client_data.buf,info->client_data.buf+i, info->client_data.buflen-i);
	info->client_data.buflen=info->client_data.buflen-i;

	if( info->client_data.buflen==0 && ~(info->state & SERVERDATA) )
		closedata(CLIENTDATA);
}

/* ------------------------------------------------------------- **
**  Flush as much buffer as we can without blocking  
** ------------------------------------------------------------- */
void writebuf2server()
{
	int i;

	debug("    Writing buffer --> server\n");
	i=iwrite(info->server_data.fd,info->server_data.buf,info->server_data.buflen);
	if(errno==EPIPE){
		closedata(SERVERDATA);
		return;
	}
	memmove(info->server_data.buf,info->server_data.buf+i, info->server_data.buflen-i);
	info->server_data.buflen=info->server_data.buflen-i;

	if( info->server_data.buflen==0 && ~(info->state & CLIENTDATA) )
		closedata(SERVERDATA);
}

/* ------------------------------------------------------------- **
**  Close data socket, and if there is nothing left to flush close
** the other one too.
** ------------------------------------------------------------- */
void closedata(int closeme)
{
	switch(closeme){
	case SERVERDATA:
		close(info->server_data.fd);
		info->state &= ~SERVERDATA;
		if(info->client_data.buflen==0) {
			close(info->client_data.fd);
			info->state &= ~CLIENTDATA;
		}
		break;
	case CLIENTDATA:
		close(info->client_data.fd);
		info->state &= ~CLIENTDATA;
		if(info->server_data.buflen==0) {
			close(info->server_data.fd);
			info->state &= ~SERVERDATA;
		}
		break;
	default:
		close(info->client_data.fd);
		close(info->server_data.fd);
		info->state &= ~CONNECTED;
	}
}












